/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "atomic.h"

Atomic::Atomic(uint64_t value)
{
	m_value = value;
	MUTEX_CREATE(m_mutex);
}

Atomic::~Atomic()
{
	MUTEX_DELETE(m_mutex);
}

uint64_t Atomic::add(uint32_t inc)
{
	uint64_t ret;

	MUTEX_LOCK(m_mutex);
	ret = m_value;
	m_value += inc;
	if (m_value < ret) // overflow
	{
		m_value = UINT64_MAX;
	}
	MUTEX_UNLOCK(m_mutex);
	return ret;
}

uint64_t Atomic::add(uint64_t inc)
{
	uint64_t ret;

	MUTEX_LOCK(m_mutex);
	ret = m_value;
	m_value += inc;
	if (m_value < ret) // overflow
	{
		m_value = UINT64_MAX;
	}
	MUTEX_UNLOCK(m_mutex);
	return ret;
}

const uint64_t Atomic::get()
{
	return m_value;
}
