/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "cudacommon.h"
#include <map>

using namespace std;

struct blocksThreads
{
	int blocks, threads;
};

map<int,blocksThreads> defaultBlocksThreads;

int versionToCores(int major, int minor)
{
	uint32_t ret = -1;

	switch ((major << 4) + minor)
	{
		case 0x10: case 0x11: case 0x12: case 0x13:
			ret = 8;
			break;

		case 0x20:
			ret = 32;
			break;

		case 0x21:
			ret = 48;
	}
	return ret;
}

void getDefaultBlocksThreads(int deviceId, int &blocks, int &threads)
{
	map<int,blocksThreads>::const_iterator it;

	it = defaultBlocksThreads.find(deviceId);
	if (it == defaultBlocksThreads.end())
	{
		cudaDeviceProp deviceProp;

		checkCudaErrors(cudaGetDeviceProperties(&deviceProp, deviceId));
		blocks  = 2 * deviceProp.multiProcessorCount;
		threads = min(deviceProp.maxThreadsPerBlock, 32 * versionToCores(deviceProp.major, deviceProp.minor));
		//threads = min(deviceProp.maxThreadsPerBlock, ((deviceProp.regsPerBlock/32) & 0xfffffff0));
	}
	else
	{
		blocks  = it->second.blocks;
		threads = it->second.threads;
	}
}

void setDefaultBlocksThreads(int deviceId, int blocks, int threads)
{
	blocksThreads bt = {blocks, threads};

	defaultBlocksThreads[deviceId] = bt;
}

void setDefaultBlocks(int deviceId, int blocks)
{
	blocksThreads bt = {blocks, 0};
	int tmp;

	getDefaultBlocksThreads(deviceId, tmp, bt.threads);
	defaultBlocksThreads[deviceId] = bt;
}

void setDefaultThreads(int deviceId, int threads)
{
	blocksThreads bt = {0, threads};
	int tmp;

	getDefaultBlocksThreads(deviceId, bt.blocks, tmp);
	defaultBlocksThreads[deviceId] = bt;
}
