/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CUDA_COMMON_H
#define CUDA_COMMON_H

#include "common.h"
#include <cuda_runtime_api.h>

#define checkCudaErrors(err) \
            __checkCudaErrors(            err,         __FILE__,       __LINE__)
inline void __checkCudaErrors(cudaError_t err, const char *file, const int line)
{
    if (cudaSuccess != err)
	{
		reportError("%s(%i) : CUDA Runtime API error %d: %s.",
			file, line, (int) err, cudaGetErrorString(err));
        exit(1);
    }
}

#define getLastCudaError(msg) \
            __getLastCudaError(            msg,         __FILE__,       __LINE__)
inline void __getLastCudaError(const char *msg, const char *file, const int line)
{
    cudaError_t err = cudaGetLastError();
    if (cudaSuccess != err)
	{
		reportError("%s(%i) : getLastCudaError() CUDA error : %s : (%d) %s.",
			file, line, msg, (int) err, cudaGetErrorString(err));
        exit(1);
    }
}

typedef void         (*kernel_doLinks)           (int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t startChainPos, uint32_t endChainPos, uint32_t numIndexes);
typedef void         (*kernel_doPreWorkLinks)    (int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t firstChainPos, uint32_t firstEndChainPos, uint32_t endChainPos, uint32_t numIndexes, const uint32_t *d_hash, bool first);
typedef unsigned int (*kernel_falseAlarmChecking)(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, const uint32_t *d_hash, const uint64_t *d_lastChainPoses, uint32_t tableIndex, uint32_t numIndexes);

int versionToCores(int major, int minor);
void getDefaultBlocksThreads(int deviceId, int &blocks, int &threads);
void setDefaultBlocksThreads(int deviceId, int  blocks, int  threads);
void setDefaultBlocks       (int deviceId, int  blocks);
void setDefaultThreads      (int deviceId,              int  threads);

#endif
