/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "divbase.h"
#include "reduxfactory.h"
#include "hashfactory.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

DivBase::~DivBase()
{
	cleanUp();
}

DivBase::DivBase(const keySpace &ks, const char *hashName)
	: ReduxBruteForce(ks, hashName)
{
	m_keySpaceLut = NULL;
	if (isInited())
	{
		// Init Lookup table for "hash64 % m_keySpace"
		uint64_t num = UINT64_C(1) << 56;
		uint64_t keySpace = m_keySpace;
		uint64_t *keySpaceLut;
		uint32_t count;

		count = (countLeadingZeros(m_keySpace) + 6) / 7;
		if (count < 2)
		{
			count = 2;
		}
		m_keySpaceLutCount = count;
		m_keySpaceLut = keySpaceLut = new uint64_t[256 * count];

		for (uint32_t i = 0; i < count; i++)
		{
			for (uint32_t j = 0; j < 256; j++, keySpaceLut++)
			{
				*keySpaceLut = ((j * num) / keySpace) * keySpace;
			}
			num >>= 7;
		}
	}
}

uint64_t DivBase::hashToIndex(const uint32_t *hash, uint32_t chainPos, uint32_t tableIndex)
{
	uint64_t index              = *((uint64_t*) hash) + (chainPos + (tableIndex << 16));
	const uint64_t *keySpaceLut = m_keySpaceLut;
	uint32_t keySpaceLutCount   = m_keySpaceLutCount;
	uint32_t shift              = 56 - 14;

	index -= keySpaceLut[ index >>  56];
	index -= keySpaceLut[(index >> (56 - 7)) + 256];
	for (uint32_t i = 2; i < keySpaceLutCount; i++)
	{
		index -= keySpaceLut[(uint32_t) (index >> shift) + 256 * i];
		shift -= 7;
	}
	if (index >= m_keySpace)
	{
		index -= m_keySpace;
	}
	return index;
}

void DivBase::cleanUp()
{
	if (m_keySpaceLut != NULL)
	{
		delete [] m_keySpaceLut;
		m_keySpaceLut = NULL;
	}

	ReduxBruteForce::cleanUp();
}

void DivBase::copy(const DivBase &r)
{
	// Don't do this because of multiple virtual inheritance
	// ReduxBruteForce::copy(r);

	m_keySpaceLut = NULL;
	if (r.m_hashFunc != NULL)
	{
		m_keySpaceLutCount = r.m_keySpaceLutCount;
		m_keySpaceLut = new uint64_t[256 * m_keySpaceLutCount];
		memcpy(m_keySpaceLut, r.m_keySpaceLut, sizeof(uint64_t) * 256 * m_keySpaceLutCount);
	}
}
