/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "generationthread.h"
#include "reduxfactory.h"

GenerationThread::GenerationThread()
{
	;
}

GenerationThread::~GenerationThread()
{
	stopRunning();
	waitForStopRunning();
}

int GenerationThread::stopRunning()
{
	m_stopRunning = 1;
	return 0;
}

void *GenerationThread::threadFunc(void *arg)
{
	uint64_t startPoint = 0, lastStartPoint, lastStartPointBlock = 0;
	Atomic *startPointAtomic;
	Redux *redux;
	uint64_t *indexes;
	WorkQueue inWorkQueue, *outWorkQueue;
	workQueueNode nodes[2], *node;
	uint32_t chainLength, tableIndex, numPws;

	m_stopRunning = 0;
	outWorkQueue     = ((generationThreadArgs*) arg)->outWorkQueue;
	startPointAtomic = ((generationThreadArgs*) arg)->startPoint;
	lastStartPoint   = ((generationThreadArgs*) arg)->lastStartPoint;
	chainLength      = ((generationThreadArgs*) arg)->chainLength;
	tableIndex       = ((generationThreadArgs*) arg)->tableIndex;
	redux = ReduxFactory::getReduxInstance( ((generationThreadArgs*) arg)->reduxName,
	                                       *((generationThreadArgs*) arg)->ks,
	                                        ((generationThreadArgs*) arg)->gpuId,
	                                        ((generationThreadArgs*) arg)->hashName);
	if (redux == NULL)
	{
		reportError("No redux found.");
		exit(1);
	}
	numPws = redux->getOptimalNumIndexes();
	if (((generationThreadArgs*) arg)->gpuId == 0)
	{
		numPws *= 128 / numPws;
	}
	assert(numPws <= CHAIN_BUFFER_LENGTH);
	if (lastStartPoint >= numPws)
	{
		lastStartPointBlock = lastStartPoint - numPws;
	}

	nodes[0].returnQueue = &inWorkQueue;
	nodes[0].data = new uint64_t[numPws + 2];
	((uint64_t*) (nodes[0].data))[0] = numPws;
	nodes[1].returnQueue = &inWorkQueue;
	nodes[1].data = new uint64_t[numPws + 2];
	((uint64_t*) (nodes[1].data))[0] = numPws;
	inWorkQueue.push(nodes + 1);
	node = nodes;
	while (m_stopRunning == 0)
	{
		startPoint = startPointAtomic->add(numPws);
		((uint64_t*) (node->data))[1] = startPoint;
		indexes = ((uint64_t*) (node->data)) + 2;
		if (startPoint >= lastStartPointBlock)
		{
			if (startPoint > lastStartPoint)
			{
				break;
			}
			numPws = (uint32_t) (lastStartPoint - startPoint + 1);
			((uint64_t*) (node->data))[0] = numPws;
			m_stopRunning = 1;
		}
		for (uint32_t i = 0; i < numPws; i++)
		{
			indexes[i] = startPoint + i;
		}
		redux->doLinks(indexes, 0, tableIndex, chainLength, numPws);
		outWorkQueue->push(node);
		node = inWorkQueue.pop();
	}
	delete redux;
	delete [] node->data;
	node = inWorkQueue.pop();
	delete [] node->data;
	return NULL;
}
