/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef REDUX_GPU_DIV_SINGLE_BYTE_H
#define REDUX_GPU_DIV_SINGLE_BYTE_H

#include "common.h"
#include "divsb.h"
#include "keyspace.h"
#include "cudacommon.h"

// Divide [Vertical Key Space]

// It's considered a vertical key space because it makes more sense on multiple lines
// aaaannn0
// aaaannn00
// aaaannn000

// currently only does the normal key spaces such as alpha#1-7 and numeric#1-12

/*
	Lookup table for "hash % keySpace"
	Cache two characters of the password
	Fix point multiply for characters
*/
class GpuDivSB : public ReduxSingleByte, public DivBase
{
public:
	virtual ~GpuDivSB();

	static Redux *create(const keySpace &ks, const char *hashName, int gpu);

	virtual void doLinks(uint64_t *indexes, uint32_t chainPos, uint32_t tableIndex, uint32_t numLinks, uint32_t numIndexes);
	virtual void doPreWorkLinks(uint64_t *indexes, uint32_t firstChainPos, uint32_t tableIndex, uint32_t chainLength, uint32_t numIndexes, const uint32_t *hash);
	virtual uint32_t falseAlarmChecking(uint64_t *indexes, const uint32_t *hash, const uint32_t *lastChainPoses, uint32_t tableIndex, uint32_t numIndexes);

private:
	GpuDivSB &operator=(const GpuDivSB &r) { return *this; }
	GpuDivSB(const keySpace &ks, const char *hashName, int gpu);
	void cleanUp();

	kernel_doLinks            m_doLinks;
	kernel_doPreWorkLinks     m_doPreWorkLinks;
	kernel_falseAlarmChecking m_falseAlarmChecking;
	uint64_t *m_d_indexes, *m_d_lastChainPoses;
	uint32_t *m_d_hash, *m_d_shared/*, m_sharedSize*/;
	int m_blocks, m_threads;

	static uint32_t initFactory();
	static uint32_t s_inited;
};

#endif
