/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef HASH_FACTORY_H
#define HASH_FACTORY_H

#include <string>
#include <list>
#include "common.h"
#include "hash.h"

/*
	MD5
	LM             DES("4B47532140232425", mid(ucase(PW) . NULL_Padding, 0, 7))
	NTLM           MD4(UTF16(PW))
	DoubleMD5      MD5(lcase(MD5(PW)))
	SHA1

	CiscoPIX       lo_3_bytes_of_each_abcd(MD5(PW))
	MySQLSHA1      SHA1(SHA1(PW, true))
	HalfMD5        get_bc(MD5(PW))

	MSCACHE        MD4(MD4(UTF16(PW), true) . UTF16(UN))                                                                                        UN = administrator
	ORACLE         right(DESCBC(UTF16BE(ucase(UN . PW)), right(DESCBC(UTF16BE(ucase(UN . PW)), "0123456789ABCDEF", vi=NULL), 8), vi=NULL), 8)   UN = SYS
	NTLMCHALL      DES(Salt, mid(NTLM(PW), 0, 7)) [. DES(Salt, mid(NTLM(PW), 7, 7))]                                                            Salt = 1122334455667788
	HalfLMChall    DES(Salt, mid(LM(PW), 0, 7))                                                                                                 Salt = 1122334455667788
	2ndHalfLMChall DES(Salt, Rand8Bits . mid(LM(PW), 0, 6)) [. DES(Salt, Rand8Bits . mid(LM(PW) . NULL_Padding, 6, 7))]                         Salt = 1122334455667788

	MD2
	MD4
	RIPEMD160
*/

struct hashInfo
{
	hashAlgorithm hash;
	std::string name, comment;
	uint32_t minPw, maxPw;
};

// AVX LM
#define MAX_PW_SIZE    256 / 4 * 3
#define MAX_HASH_SIZE  256 / 4 * 2
// Double MD5
#define MAX_PW_BLOCKS  9

class HashFactory
{
public:
	static uint32_t getHashCharacterEncoding(const char *name);
	static uint32_t getHashInstance         (const char *name, uint32_t minPw, uint32_t maxPw, hashAlgorithm &hash);
	static uint32_t registerHash            (const char *name, const char *comment, uint32_t minPw, uint32_t maxPw, HASH_FUNC hashFunc, uint32_t hashLen, uint32_t hashesPerCall, uint32_t characterEncoding, uint32_t numPwBlocks, uint32_t numExtraHashBlocks = 0, uint32_t pwByteSkip = 0);
	static void     printHashes();

private:
	HashFactory() {}
	static std::list<hashInfo*> *s_hashes;
};

#endif
