/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "hashfactory.h"
#include "md5.h"
#include "md5help.h"

uint32_t MD5::inited = MD5::initCalc();

#ifdef ARC_x86

void MD5::calcSSE2(vector4 *pw, vector4 *hash)
{
	md5Func((__m128i*) pw, (__m128i*) hash);
}

#endif

void MD5::calcNorm(vector4 *pw, vector4 *hash)
{
	uint32_t a = 0x67452301, b = 0xefcdab89, c = 0x98badcfe, d = 0x10325476;

	#define MD5_FF(a, b, c, d, C, x, s) a += (d ^ (b & (c ^ d))) + C + FROM_LITTLE_ENDIAN_32(x); a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_GG(a, b, c, d, C, x, s) a += (c ^ (d & (b ^ c))) + C + FROM_LITTLE_ENDIAN_32(x); a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_HH(a, b, c, d, C, x, s) a += (b ^ c ^ d)         + C + FROM_LITTLE_ENDIAN_32(x); a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_II(a, b, c, d, C, x, s) a += (c ^ (b | ~d))      + C + FROM_LITTLE_ENDIAN_32(x); a = (a << s) | (a >> (32 - s)); a += b;

	#define x0   pw[1].data[0]
	#define x1   pw[2].data[0]
	#define x2   pw[3].data[0]
	#define x3   pw[4].data[0]
	#define x4   pw[5].data[0]
	#define x5   0
	#define x6   0
	#define x7   0
	#define x8   0
	#define x9   0
	#define x10  0
	#define x11  0
	#define x12  0
	#define x13  0
	#define x14  pw[0].data[0]
	#define x15  0

	pw[0].data[0] <<= 3;

	MD5_FF(a, b, c, d, 0xd76aa478,  x0,  7); MD5_FF(d, a, b, c, 0xe8c7b756,  x1, 12); MD5_FF(c, d, a, b, 0x242070db,  x2, 17); MD5_FF(b, c, d, a, 0xc1bdceee,  x3, 22);
	MD5_FF(a, b, c, d, 0xf57c0faf,  x4,  7); MD5_FF(d, a, b, c, 0x4787c62a,  x5, 12); MD5_FF(c, d, a, b, 0xa8304613,  x6, 17); MD5_FF(b, c, d, a, 0xfd469501,  x7, 22);
	MD5_FF(a, b, c, d, 0x698098d8,  x8,  7); MD5_FF(d, a, b, c, 0x8b44f7af,  x9, 12); MD5_FF(c, d, a, b, 0xffff5bb1, x10, 17); MD5_FF(b, c, d, a, 0x895cd7be, x11, 22);
	MD5_FF(a, b, c, d, 0x6b901122, x12,  7); MD5_FF(d, a, b, c, 0xfd987193, x13, 12); MD5_FF(c, d, a, b, 0xa679438e, x14, 17); MD5_FF(b, c, d, a, 0x49b40821, x15, 22);

	MD5_GG(a, b, c, d, 0xf61e2562,  x1,  5); MD5_GG(d, a, b, c, 0xc040b340,  x6,  9); MD5_GG(c, d, a, b, 0x265e5a51, x11, 14); MD5_GG(b, c, d, a, 0xe9b6c7aa,  x0, 20);
	MD5_GG(a, b, c, d, 0xd62f105d,  x5,  5); MD5_GG(d, a, b, c, 0x02441453, x10,  9); MD5_GG(c, d, a, b, 0xd8a1e681, x15, 14); MD5_GG(b, c, d, a, 0xe7d3fbc8,  x4, 20);
	MD5_GG(a, b, c, d, 0x21e1cde6,  x9,  5); MD5_GG(d, a, b, c, 0xc33707d6, x14,  9); MD5_GG(c, d, a, b, 0xf4d50d87,  x3, 14); MD5_GG(b, c, d, a, 0x455a14ed,  x8, 20);
	MD5_GG(a, b, c, d, 0xa9e3e905, x13,  5); MD5_GG(d, a, b, c, 0xfcefa3f8,  x2,  9); MD5_GG(c, d, a, b, 0x676f02d9,  x7, 14); MD5_GG(b, c, d, a, 0x8d2a4c8a, x12, 20);

	MD5_HH(a, b, c, d, 0xfffa3942,  x5,  4); MD5_HH(d, a, b, c, 0x8771f681,  x8, 11); MD5_HH(c, d, a, b, 0x6d9d6122, x11, 16); MD5_HH(b, c, d, a, 0xfde5380c, x14, 23);
	MD5_HH(a, b, c, d, 0xa4beea44,  x1,  4); MD5_HH(d, a, b, c, 0x4bdecfa9,  x4, 11); MD5_HH(c, d, a, b, 0xf6bb4b60,  x7, 16); MD5_HH(b, c, d, a, 0xbebfbc70, x10, 23);
	MD5_HH(a, b, c, d, 0x289b7ec6, x13,  4); MD5_HH(d, a, b, c, 0xeaa127fa,  x0, 11); MD5_HH(c, d, a, b, 0xd4ef3085,  x3, 16); MD5_HH(b, c, d, a, 0x04881d05,  x6, 23);
	MD5_HH(a, b, c, d, 0xd9d4d039,  x9,  4); MD5_HH(d, a, b, c, 0xe6db99e5, x12, 11); MD5_HH(c, d, a, b, 0x1fa27cf8, x15, 16); MD5_HH(b, c, d, a, 0xc4ac5665,  x2, 23);

	MD5_II(a, b, c, d, 0xf4292244,  x0,  6); MD5_II(d, a, b, c, 0x432aff97,  x7, 10); MD5_II(c, d, a, b, 0xab9423a7, x14, 15); MD5_II(b, c, d, a, 0xfc93a039,  x5, 21);
	MD5_II(a, b, c, d, 0x655b59c3, x12,  6); MD5_II(d, a, b, c, 0x8f0ccc92,  x3, 10); MD5_II(c, d, a, b, 0xffeff47d, x10, 15); MD5_II(b, c, d, a, 0x85845dd1,  x1, 21);
	MD5_II(a, b, c, d, 0x6fa87e4f,  x8,  6); MD5_II(d, a, b, c, 0xfe2ce6e0, x15, 10); MD5_II(c, d, a, b, 0xa3014314,  x6, 15); MD5_II(b, c, d, a, 0x4e0811a1, x13, 21);
	MD5_II(a, b, c, d, 0xf7537e82,  x4,  6); MD5_II(d, a, b, c, 0xbd3af235, x11, 10); MD5_II(c, d, a, b, 0x2ad7d2bb,  x2, 15); MD5_II(b, c, d, a, 0xeb86d391,  x9, 21);

	#undef MD5_FF
	#undef MD5_GG
	#undef MD5_HH
	#undef MD5_II

	#undef x0
	#undef x1
	#undef x2
	#undef x3
	#undef x4
	#undef x5
	#undef x6
	#undef x7
	#undef x8
	#undef x9
	#undef x10
	#undef x11
	#undef x12
	#undef x13
	#undef x14
	#undef x15

	hash[0].data[0] = TO_LITTLE_ENDIAN_32(a + 0x67452301);
	hash[1].data[0] = TO_LITTLE_ENDIAN_32(b + 0xefcdab89);
	hash[2].data[0] = TO_LITTLE_ENDIAN_32(c + 0x98badcfe);
	hash[3].data[0] = TO_LITTLE_ENDIAN_32(d + 0x10325476);
}

uint32_t MD5::initCalc()
{
#ifdef ARC_x86
	if ((getInstructionSets() & (IS_AVX | IS_XOP)) == (IS_AVX | IS_XOP))
	{
		return HashFactory::registerHash("md5", QUOTE(__MD5_INTERLACED__) "xAVX_XOP", 0, 19, MD5::calcAVXXOP,  16, 4 * __MD5_INTERLACED__, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 5);
	}
	else if ((getInstructionSets() & IS_AVX) == IS_AVX)
	{
		return HashFactory::registerHash("md5", QUOTE(__MD5_INTERLACED__) "xAVX",     0, 19, MD5::calcAVX,     16, 4 * __MD5_INTERLACED__, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 5);
	}
	else if ((getInstructionSets() & IS_SSE2) == IS_SSE2)
	{
		return HashFactory::registerHash("md5", QUOTE(__MD5_INTERLACED__) "xSSE2",    0, 19, MD5::calcSSE2,    16, 4 * __MD5_INTERLACED__, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 5);
	}
#endif
	return HashFactory::registerHash("md5", "scalar",                             0, 19, MD5::calcNorm,    16,                      1, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 5);
}
