/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MD5_HELP_H
#define MD5_HELP_H

#include "common.h"

#ifndef ARC_x86
	// Non-x86... Yeah I got nothing
#else

#ifdef _WIN32
	#include <immintrin.h> // AVX and older
	#include <ammintrin.h> // XOP
#else
	#include <x86intrin.h>
#endif

#ifdef ARC_x86_64
	#define __MD5_INTERLACED__ 3
#elif defined(ARC_x86_32)
	#define __MD5_INTERLACED__ 3
#endif

#if __MD5_INTERLACED__ > 4
	#undef  __MD5_INTERLACED__
	#define __MD5_INTERLACED__ 4
#endif

#if defined(AVX_BROADCAST) && defined(AVX_CODE)
	#define LOAD_CONST_MD5_A     _mm_broadcastd_epi32(0x67452301)
	#define LOAD_CONST_MD5_B     _mm_broadcastd_epi32(0xefcdab89)
	#define LOAD_CONST_MD5_C     _mm_broadcastd_epi32(0x98badcfe)
	#define LOAD_CONST_MD5_D     _mm_broadcastd_epi32(0x10325476)

	#define LOAD_CONST_MD5_FF0   _mm_broadcastd_epi32(0xd76aa478)
	#define LOAD_CONST_MD5_FF1   _mm_broadcastd_epi32(0xe8c7b756)
	#define LOAD_CONST_MD5_FF2   _mm_broadcastd_epi32(0x242070db)
	#define LOAD_CONST_MD5_FF3   _mm_broadcastd_epi32(0xc1bdceee)
	#define LOAD_CONST_MD5_FF4   _mm_broadcastd_epi32(0xf57c0faf)
	#define LOAD_CONST_MD5_FF5   _mm_broadcastd_epi32(0x4787c62a)
	#define LOAD_CONST_MD5_FF6   _mm_broadcastd_epi32(0xa8304613)
	#define LOAD_CONST_MD5_FF7   _mm_broadcastd_epi32(0xfd469501)
	#define LOAD_CONST_MD5_FF8   _mm_broadcastd_epi32(0x698098d8)
	#define LOAD_CONST_MD5_FF9   _mm_broadcastd_epi32(0x8b44f7af)
	#define LOAD_CONST_MD5_FF10  _mm_broadcastd_epi32(0xffff5bb1)
	#define LOAD_CONST_MD5_FF11  _mm_broadcastd_epi32(0x895cd7be)
	#define LOAD_CONST_MD5_FF12  _mm_broadcastd_epi32(0x6b901122)
	#define LOAD_CONST_MD5_FF13  _mm_broadcastd_epi32(0xfd987193)
	#define LOAD_CONST_MD5_FF14  _mm_broadcastd_epi32(0xa679438e)
	#define LOAD_CONST_MD5_FF15  _mm_broadcastd_epi32(0x49b40821)

	#define LOAD_CONST_MD5_GG0   _mm_broadcastd_epi32(0xf61e2562)
	#define LOAD_CONST_MD5_GG1   _mm_broadcastd_epi32(0xc040b340)
	#define LOAD_CONST_MD5_GG2   _mm_broadcastd_epi32(0x265e5a51)
	#define LOAD_CONST_MD5_GG3   _mm_broadcastd_epi32(0xe9b6c7aa)
	#define LOAD_CONST_MD5_GG4   _mm_broadcastd_epi32(0xd62f105d)
	#define LOAD_CONST_MD5_GG5   _mm_broadcastd_epi32(0x02441453)
	#define LOAD_CONST_MD5_GG6   _mm_broadcastd_epi32(0xd8a1e681)
	#define LOAD_CONST_MD5_GG7   _mm_broadcastd_epi32(0xe7d3fbc8)
	#define LOAD_CONST_MD5_GG8   _mm_broadcastd_epi32(0x21e1cde6)
	#define LOAD_CONST_MD5_GG9   _mm_broadcastd_epi32(0xc33707d6)
	#define LOAD_CONST_MD5_GG10  _mm_broadcastd_epi32(0xf4d50d87)
	#define LOAD_CONST_MD5_GG11  _mm_broadcastd_epi32(0x455a14ed)
	#define LOAD_CONST_MD5_GG12  _mm_broadcastd_epi32(0xa9e3e905)
	#define LOAD_CONST_MD5_GG13  _mm_broadcastd_epi32(0xfcefa3f8)
	#define LOAD_CONST_MD5_GG14  _mm_broadcastd_epi32(0x676f02d9)
	#define LOAD_CONST_MD5_GG15  _mm_broadcastd_epi32(0x8d2a4c8a)

	#define LOAD_CONST_MD5_HH0   _mm_broadcastd_epi32(0xfffa3942)
	#define LOAD_CONST_MD5_HH1   _mm_broadcastd_epi32(0x8771f681)
	#define LOAD_CONST_MD5_HH2   _mm_broadcastd_epi32(0x6d9d6122)
	#define LOAD_CONST_MD5_HH3   _mm_broadcastd_epi32(0xfde5380c)
	#define LOAD_CONST_MD5_HH4   _mm_broadcastd_epi32(0xa4beea44)
	#define LOAD_CONST_MD5_HH5   _mm_broadcastd_epi32(0x4bdecfa9)
	#define LOAD_CONST_MD5_HH6   _mm_broadcastd_epi32(0xf6bb4b60)
	#define LOAD_CONST_MD5_HH7   _mm_broadcastd_epi32(0xbebfbc70)
	#define LOAD_CONST_MD5_HH8   _mm_broadcastd_epi32(0x289b7ec6)
	#define LOAD_CONST_MD5_HH9   _mm_broadcastd_epi32(0xeaa127fa)
	#define LOAD_CONST_MD5_HH10  _mm_broadcastd_epi32(0xd4ef3085)
	#define LOAD_CONST_MD5_HH11  _mm_broadcastd_epi32(0x04881d05)
	#define LOAD_CONST_MD5_HH12  _mm_broadcastd_epi32(0xd9d4d039)
	#define LOAD_CONST_MD5_HH13  _mm_broadcastd_epi32(0xe6db99e5)
	#define LOAD_CONST_MD5_HH14  _mm_broadcastd_epi32(0x1fa27cf8)
	#define LOAD_CONST_MD5_HH15  _mm_broadcastd_epi32(0xc4ac5665)

	#define LOAD_CONST_MD5_II0   _mm_broadcastd_epi32(0xf4292244)
	#define LOAD_CONST_MD5_II1   _mm_broadcastd_epi32(0x432aff97)
	#define LOAD_CONST_MD5_II2   _mm_broadcastd_epi32(0xab9423a7)
	#define LOAD_CONST_MD5_II3   _mm_broadcastd_epi32(0xfc93a039)
	#define LOAD_CONST_MD5_II4   _mm_broadcastd_epi32(0x655b59c3)
	#define LOAD_CONST_MD5_II5   _mm_broadcastd_epi32(0x8f0ccc92)
	#define LOAD_CONST_MD5_II6   _mm_broadcastd_epi32(0xffeff47d)
	#define LOAD_CONST_MD5_II7   _mm_broadcastd_epi32(0x85845dd1)
	#define LOAD_CONST_MD5_II8   _mm_broadcastd_epi32(0x6fa87e4f)
	#define LOAD_CONST_MD5_II9   _mm_broadcastd_epi32(0xfe2ce6e0)
	#define LOAD_CONST_MD5_II10  _mm_broadcastd_epi32(0xa3014314)
	#define LOAD_CONST_MD5_II11  _mm_broadcastd_epi32(0x4e0811a1)
	#define LOAD_CONST_MD5_II12  _mm_broadcastd_epi32(0xf7537e82)
	#define LOAD_CONST_MD5_II13  _mm_broadcastd_epi32(0xbd3af235)
	#define LOAD_CONST_MD5_II14  _mm_broadcastd_epi32(0x2ad7d2bb)
	#define LOAD_CONST_MD5_II15  _mm_broadcastd_epi32(0xeb86d391)
#else
	#define LOAD_CONST_MD5_A     _mm_set_epi32(0x67452301,0x67452301,0x67452301,0x67452301)
	#define LOAD_CONST_MD5_B     _mm_set_epi32(0xefcdab89,0xefcdab89,0xefcdab89,0xefcdab89)
	#define LOAD_CONST_MD5_C     _mm_set_epi32(0x98badcfe,0x98badcfe,0x98badcfe,0x98badcfe)
	#define LOAD_CONST_MD5_D     _mm_set_epi32(0x10325476,0x10325476,0x10325476,0x10325476)

	#define LOAD_CONST_MD5_FF0   _mm_set_epi32(0xd76aa478,0xd76aa478,0xd76aa478,0xd76aa478)
	#define LOAD_CONST_MD5_FF1   _mm_set_epi32(0xe8c7b756,0xe8c7b756,0xe8c7b756,0xe8c7b756)
	#define LOAD_CONST_MD5_FF2   _mm_set_epi32(0x242070db,0x242070db,0x242070db,0x242070db)
	#define LOAD_CONST_MD5_FF3   _mm_set_epi32(0xc1bdceee,0xc1bdceee,0xc1bdceee,0xc1bdceee)
	#define LOAD_CONST_MD5_FF4   _mm_set_epi32(0xf57c0faf,0xf57c0faf,0xf57c0faf,0xf57c0faf)
	#define LOAD_CONST_MD5_FF5   _mm_set_epi32(0x4787c62a,0x4787c62a,0x4787c62a,0x4787c62a)
	#define LOAD_CONST_MD5_FF6   _mm_set_epi32(0xa8304613,0xa8304613,0xa8304613,0xa8304613)
	#define LOAD_CONST_MD5_FF7   _mm_set_epi32(0xfd469501,0xfd469501,0xfd469501,0xfd469501)
	#define LOAD_CONST_MD5_FF8   _mm_set_epi32(0x698098d8,0x698098d8,0x698098d8,0x698098d8)
	#define LOAD_CONST_MD5_FF9   _mm_set_epi32(0x8b44f7af,0x8b44f7af,0x8b44f7af,0x8b44f7af)
	#define LOAD_CONST_MD5_FF10  _mm_set_epi32(0xffff5bb1,0xffff5bb1,0xffff5bb1,0xffff5bb1)
	#define LOAD_CONST_MD5_FF11  _mm_set_epi32(0x895cd7be,0x895cd7be,0x895cd7be,0x895cd7be)
	#define LOAD_CONST_MD5_FF12  _mm_set_epi32(0x6b901122,0x6b901122,0x6b901122,0x6b901122)
	#define LOAD_CONST_MD5_FF13  _mm_set_epi32(0xfd987193,0xfd987193,0xfd987193,0xfd987193)
	#define LOAD_CONST_MD5_FF14  _mm_set_epi32(0xa679438e,0xa679438e,0xa679438e,0xa679438e)
	#define LOAD_CONST_MD5_FF15  _mm_set_epi32(0x49b40821,0x49b40821,0x49b40821,0x49b40821)

	#define LOAD_CONST_MD5_GG0   _mm_set_epi32(0xf61e2562,0xf61e2562,0xf61e2562,0xf61e2562)
	#define LOAD_CONST_MD5_GG1   _mm_set_epi32(0xc040b340,0xc040b340,0xc040b340,0xc040b340)
	#define LOAD_CONST_MD5_GG2   _mm_set_epi32(0x265e5a51,0x265e5a51,0x265e5a51,0x265e5a51)
	#define LOAD_CONST_MD5_GG3   _mm_set_epi32(0xe9b6c7aa,0xe9b6c7aa,0xe9b6c7aa,0xe9b6c7aa)
	#define LOAD_CONST_MD5_GG4   _mm_set_epi32(0xd62f105d,0xd62f105d,0xd62f105d,0xd62f105d)
	#define LOAD_CONST_MD5_GG5   _mm_set_epi32(0x02441453,0x02441453,0x02441453,0x02441453)
	#define LOAD_CONST_MD5_GG6   _mm_set_epi32(0xd8a1e681,0xd8a1e681,0xd8a1e681,0xd8a1e681)
	#define LOAD_CONST_MD5_GG7   _mm_set_epi32(0xe7d3fbc8,0xe7d3fbc8,0xe7d3fbc8,0xe7d3fbc8)
	#define LOAD_CONST_MD5_GG8   _mm_set_epi32(0x21e1cde6,0x21e1cde6,0x21e1cde6,0x21e1cde6)
	#define LOAD_CONST_MD5_GG9   _mm_set_epi32(0xc33707d6,0xc33707d6,0xc33707d6,0xc33707d6)
	#define LOAD_CONST_MD5_GG10  _mm_set_epi32(0xf4d50d87,0xf4d50d87,0xf4d50d87,0xf4d50d87)
	#define LOAD_CONST_MD5_GG11  _mm_set_epi32(0x455a14ed,0x455a14ed,0x455a14ed,0x455a14ed)
	#define LOAD_CONST_MD5_GG12  _mm_set_epi32(0xa9e3e905,0xa9e3e905,0xa9e3e905,0xa9e3e905)
	#define LOAD_CONST_MD5_GG13  _mm_set_epi32(0xfcefa3f8,0xfcefa3f8,0xfcefa3f8,0xfcefa3f8)
	#define LOAD_CONST_MD5_GG14  _mm_set_epi32(0x676f02d9,0x676f02d9,0x676f02d9,0x676f02d9)
	#define LOAD_CONST_MD5_GG15  _mm_set_epi32(0x8d2a4c8a,0x8d2a4c8a,0x8d2a4c8a,0x8d2a4c8a)

	#define LOAD_CONST_MD5_HH0   _mm_set_epi32(0xfffa3942,0xfffa3942,0xfffa3942,0xfffa3942)
	#define LOAD_CONST_MD5_HH1   _mm_set_epi32(0x8771f681,0x8771f681,0x8771f681,0x8771f681)
	#define LOAD_CONST_MD5_HH2   _mm_set_epi32(0x6d9d6122,0x6d9d6122,0x6d9d6122,0x6d9d6122)
	#define LOAD_CONST_MD5_HH3   _mm_set_epi32(0xfde5380c,0xfde5380c,0xfde5380c,0xfde5380c)
	#define LOAD_CONST_MD5_HH4   _mm_set_epi32(0xa4beea44,0xa4beea44,0xa4beea44,0xa4beea44)
	#define LOAD_CONST_MD5_HH5   _mm_set_epi32(0x4bdecfa9,0x4bdecfa9,0x4bdecfa9,0x4bdecfa9)
	#define LOAD_CONST_MD5_HH6   _mm_set_epi32(0xf6bb4b60,0xf6bb4b60,0xf6bb4b60,0xf6bb4b60)
	#define LOAD_CONST_MD5_HH7   _mm_set_epi32(0xbebfbc70,0xbebfbc70,0xbebfbc70,0xbebfbc70)
	#define LOAD_CONST_MD5_HH8   _mm_set_epi32(0x289b7ec6,0x289b7ec6,0x289b7ec6,0x289b7ec6)
	#define LOAD_CONST_MD5_HH9   _mm_set_epi32(0xeaa127fa,0xeaa127fa,0xeaa127fa,0xeaa127fa)
	#define LOAD_CONST_MD5_HH10  _mm_set_epi32(0xd4ef3085,0xd4ef3085,0xd4ef3085,0xd4ef3085)
	#define LOAD_CONST_MD5_HH11  _mm_set_epi32(0x04881d05,0x04881d05,0x04881d05,0x04881d05)
	#define LOAD_CONST_MD5_HH12  _mm_set_epi32(0xd9d4d039,0xd9d4d039,0xd9d4d039,0xd9d4d039)
	#define LOAD_CONST_MD5_HH13  _mm_set_epi32(0xe6db99e5,0xe6db99e5,0xe6db99e5,0xe6db99e5)
	#define LOAD_CONST_MD5_HH14  _mm_set_epi32(0x1fa27cf8,0x1fa27cf8,0x1fa27cf8,0x1fa27cf8)
	#define LOAD_CONST_MD5_HH15  _mm_set_epi32(0xc4ac5665,0xc4ac5665,0xc4ac5665,0xc4ac5665)

	#define LOAD_CONST_MD5_II0   _mm_set_epi32(0xf4292244,0xf4292244,0xf4292244,0xf4292244)
	#define LOAD_CONST_MD5_II1   _mm_set_epi32(0x432aff97,0x432aff97,0x432aff97,0x432aff97)
	#define LOAD_CONST_MD5_II2   _mm_set_epi32(0xab9423a7,0xab9423a7,0xab9423a7,0xab9423a7)
	#define LOAD_CONST_MD5_II3   _mm_set_epi32(0xfc93a039,0xfc93a039,0xfc93a039,0xfc93a039)
	#define LOAD_CONST_MD5_II4   _mm_set_epi32(0x655b59c3,0x655b59c3,0x655b59c3,0x655b59c3)
	#define LOAD_CONST_MD5_II5   _mm_set_epi32(0x8f0ccc92,0x8f0ccc92,0x8f0ccc92,0x8f0ccc92)
	#define LOAD_CONST_MD5_II6   _mm_set_epi32(0xffeff47d,0xffeff47d,0xffeff47d,0xffeff47d)
	#define LOAD_CONST_MD5_II7   _mm_set_epi32(0x85845dd1,0x85845dd1,0x85845dd1,0x85845dd1)
	#define LOAD_CONST_MD5_II8   _mm_set_epi32(0x6fa87e4f,0x6fa87e4f,0x6fa87e4f,0x6fa87e4f)
	#define LOAD_CONST_MD5_II9   _mm_set_epi32(0xfe2ce6e0,0xfe2ce6e0,0xfe2ce6e0,0xfe2ce6e0)
	#define LOAD_CONST_MD5_II10  _mm_set_epi32(0xa3014314,0xa3014314,0xa3014314,0xa3014314)
	#define LOAD_CONST_MD5_II11  _mm_set_epi32(0x4e0811a1,0x4e0811a1,0x4e0811a1,0x4e0811a1)
	#define LOAD_CONST_MD5_II12  _mm_set_epi32(0xf7537e82,0xf7537e82,0xf7537e82,0xf7537e82)
	#define LOAD_CONST_MD5_II13  _mm_set_epi32(0xbd3af235,0xbd3af235,0xbd3af235,0xbd3af235)
	#define LOAD_CONST_MD5_II14  _mm_set_epi32(0x2ad7d2bb,0x2ad7d2bb,0x2ad7d2bb,0x2ad7d2bb)
	#define LOAD_CONST_MD5_II15  _mm_set_epi32(0xeb86d391,0xeb86d391,0xeb86d391,0xeb86d391)
#endif

#define ZERO     _mm_set_epi32(0,0,0,0)

#define MD5_FF_(    a,b,c,d,x,i,j,s)    if (__MD5_INTERLACED__ > j) { a[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(a[j], _mm_or_si128(_mm_and_si128(b[j], c[j]), _mm_andnot_si128(b[j], d[j]))), x(j)), LOAD_CONST_MD5_FF##i); a[j] = _mm_add_epi32(ROT(a[j], s), b[j]); }
#define MD5_GG_(    a,b,c,d,x,i,j,s)    if (__MD5_INTERLACED__ > j) { a[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(a[j], _mm_or_si128(_mm_and_si128(b[j], d[j]), _mm_andnot_si128(d[j], c[j]))), x(j)), LOAD_CONST_MD5_GG##i); a[j] = _mm_add_epi32(ROT(a[j], s), b[j]); }
#define MD5_HH_(    a,b,c,d,x,i,j,s)    if (__MD5_INTERLACED__ > j) { a[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(a[j], _mm_xor_si128(b[j], _mm_xor_si128(c[j], d[j]))                       ), x(j)), LOAD_CONST_MD5_HH##i); a[j] = _mm_add_epi32(ROT(a[j], s), b[j]); }
#define MD5_HH_E_(  a,b,c,d,x,i,j,s,t)  if (__MD5_INTERLACED__ > j) { t[j] = _mm_xor_si128(b[j], c[j]); a[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(a[j], _mm_xor_si128(t[j], d[j])          ), x(j)), LOAD_CONST_MD5_HH##i); a[j] = _mm_add_epi32(ROT(a[j], s), b[j]); }
#define MD5_HH_O_(  a,b,c,d,x,i,j,s,t)  if (__MD5_INTERLACED__ > j) { a[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(a[j], _mm_xor_si128(b[j], t[j])                                            ), x(j)), LOAD_CONST_MD5_HH##i); a[j] = _mm_add_epi32(ROT(a[j], s), b[j]); }
#define MD5_HH16_(  a,b,c,d,x,i,j,s)    if (__MD5_INTERLACED__ > j) { a[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(a[j], _mm_xor_si128(b[j], _mm_xor_si128(c[j], d[j]))                       ), x(j)), LOAD_CONST_MD5_HH##i); a[j] = _mm_add_epi32(ROT16(a[j]) , b[j]); }
#define MD5_HH16_E_(a,b,c,d,x,i,j,s,t)  if (__MD5_INTERLACED__ > j) { t[j] = _mm_xor_si128(b[j], c[j]); a[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(a[j], _mm_xor_si128(t[j], d[j])          ), x(j)), LOAD_CONST_MD5_HH##i); a[j] = _mm_add_epi32(ROT16(a[j]) , b[j]); }
#define MD5_II_(    a,b,c,d,x,i,j,s,n)  if (__MD5_INTERLACED__ > j) { a[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(a[j], _mm_xor_si128(c[j], _mm_or_si128(b[j], _mm_xor_si128(d[j], n)))      ), x(j)), LOAD_CONST_MD5_II##i); a[j] = _mm_add_epi32(ROT(a[j], s), b[j]); }

#define MD5_FF(    a,b,c,d,x,i,s)       MD5_FF_(    a,b,c,d,x,i,0,s);   MD5_FF_(    a,b,c,d,x,i,1,s);   MD5_FF_(    a,b,c,d,x,i,2,s);   MD5_FF_(    a,b,c,d,x,i,3,s);
#define MD5_GG(    a,b,c,d,x,i,s)       MD5_GG_(    a,b,c,d,x,i,0,s);   MD5_GG_(    a,b,c,d,x,i,1,s);   MD5_GG_(    a,b,c,d,x,i,2,s);   MD5_GG_(    a,b,c,d,x,i,3,s);
#define MD5_HH(    a,b,c,d,x,i,s)       MD5_HH_(    a,b,c,d,x,i,0,s);   MD5_HH_(    a,b,c,d,x,i,1,s);   MD5_HH_(    a,b,c,d,x,i,2,s);   MD5_HH_(    a,b,c,d,x,i,3,s);
#define MD5_HH_E(  a,b,c,d,x,i,s,t)     MD5_HH_E_(  a,b,c,d,x,i,0,s,t); MD5_HH_E_(  a,b,c,d,x,i,1,s,t); MD5_HH_E_(  a,b,c,d,x,i,2,s,t); MD5_HH_E_(  a,b,c,d,x,i,3,s,t);
#define MD5_HH_O(  a,b,c,d,x,i,s,t)     MD5_HH_O_(  a,b,c,d,x,i,0,s,t); MD5_HH_O_(  a,b,c,d,x,i,1,s,t); MD5_HH_O_(  a,b,c,d,x,i,2,s,t); MD5_HH_O_(  a,b,c,d,x,i,3,s,t);
#define MD5_HH16(  a,b,c,d,x,i,s)       MD5_HH16_(  a,b,c,d,x,i,0,s);   MD5_HH16_(  a,b,c,d,x,i,1,s);   MD5_HH16_(  a,b,c,d,x,i,2,s);   MD5_HH16_(  a,b,c,d,x,i,3,s);
#define MD5_HH16_E(a,b,c,d,x,i,s,t)     MD5_HH16_E_(a,b,c,d,x,i,0,s,t); MD5_HH16_E_(a,b,c,d,x,i,1,s,t); MD5_HH16_E_(a,b,c,d,x,i,2,s,t); MD5_HH16_E_(a,b,c,d,x,i,3,s,t);
#define MD5_II(    a,b,c,d,x,i,s,n)     MD5_II_(    a,b,c,d,x,i,0,s,n); MD5_II_(    a,b,c,d,x,i,1,s,n); MD5_II_(    a,b,c,d,x,i,2,s,n); MD5_II_(    a,b,c,d,x,i,3,s,n);

#define x0(i)   pw[1 * __MD5_INTERLACED__ + i]
#define x1(i)   pw[2 * __MD5_INTERLACED__ + i]
#define x2(i)   pw[3 * __MD5_INTERLACED__ + i]
#define x3(i)   pw[4 * __MD5_INTERLACED__ + i]
#define x4(i)   pw[5 * __MD5_INTERLACED__ + i]
#define x5(i)   ZERO
#define x6(i)   ZERO
#define x7(i)   ZERO
#define x8(i)   ZERO
#define x9(i)   ZERO
#define x10(i)  ZERO
#define x11(i)  ZERO
#define x12(i)  ZERO
#define x13(i)  ZERO
#define x14(i)  pw[0 * __MD5_INTERLACED__ + i]
#define x15(i)  ZERO

#define ROT(n,s) _mm_or_si128(_mm_slli_epi32(n, s), _mm_srli_epi32(n, (32 - s)))
#define ROT16(n) _mm_shufflelo_epi16(_mm_shufflehi_epi16(n, 0xb1), 0xb1)
__inline__ void md5Func(__m128i *pw, __m128i *hash)
{
	__m128i a[__MD5_INTERLACED__], b[__MD5_INTERLACED__], c[__MD5_INTERLACED__], d[__MD5_INTERLACED__];

	if (__MD5_INTERLACED__ > 0) { a[0] = LOAD_CONST_MD5_A; }
	if (__MD5_INTERLACED__ > 1) { a[1] = LOAD_CONST_MD5_A; }
	if (__MD5_INTERLACED__ > 2) { a[2] = LOAD_CONST_MD5_A; }
	if (__MD5_INTERLACED__ > 3) { a[3] = LOAD_CONST_MD5_A; }
	if (__MD5_INTERLACED__ > 0) { b[0] = LOAD_CONST_MD5_B; }
	if (__MD5_INTERLACED__ > 1) { b[1] = LOAD_CONST_MD5_B; }
	if (__MD5_INTERLACED__ > 2) { b[2] = LOAD_CONST_MD5_B; }
	if (__MD5_INTERLACED__ > 3) { b[3] = LOAD_CONST_MD5_B; }
	if (__MD5_INTERLACED__ > 0) { c[0] = LOAD_CONST_MD5_C; }
	if (__MD5_INTERLACED__ > 1) { c[1] = LOAD_CONST_MD5_C; }
	if (__MD5_INTERLACED__ > 2) { c[2] = LOAD_CONST_MD5_C; }
	if (__MD5_INTERLACED__ > 3) { c[3] = LOAD_CONST_MD5_C; }
	if (__MD5_INTERLACED__ > 0) { d[0] = LOAD_CONST_MD5_D; }
	if (__MD5_INTERLACED__ > 1) { d[1] = LOAD_CONST_MD5_D; }
	if (__MD5_INTERLACED__ > 2) { d[2] = LOAD_CONST_MD5_D; }
	if (__MD5_INTERLACED__ > 3) { d[3] = LOAD_CONST_MD5_D; }

	if (__MD5_INTERLACED__ > 0) { x14(0) = _mm_slli_epi32(x14(0), 3); }
	if (__MD5_INTERLACED__ > 1) { x14(1) = _mm_slli_epi32(x14(1), 3); }
	if (__MD5_INTERLACED__ > 2) { x14(2) = _mm_slli_epi32(x14(2), 3); }
	if (__MD5_INTERLACED__ > 3) { x14(3) = _mm_slli_epi32(x14(3), 3); }

	MD5_FF  (a,b,c,d, x0, 0, 7    ); MD5_FF  (d,a,b,c, x1, 1,12    ); MD5_FF    (c,d,a,b, x2, 2,17    ); MD5_FF  (b,c,d,a, x3, 3,22    );
	MD5_FF  (a,b,c,d, x4, 4, 7    ); MD5_FF  (d,a,b,c, x5, 5,12    ); MD5_FF    (c,d,a,b, x6, 6,17    ); MD5_FF  (b,c,d,a, x7, 7,22    );
	MD5_FF  (a,b,c,d, x8, 8, 7    ); MD5_FF  (d,a,b,c, x9, 9,12    ); MD5_FF    (c,d,a,b,x10,10,17    ); MD5_FF  (b,c,d,a,x11,11,22    );
	MD5_FF  (a,b,c,d,x12,12, 7    ); MD5_FF  (d,a,b,c,x13,13,12    ); MD5_FF    (c,d,a,b,x14,14,17    ); MD5_FF  (b,c,d,a,x15,15,22    );

	MD5_GG  (a,b,c,d, x1, 0, 5    ); MD5_GG  (d,a,b,c, x6, 1, 9    ); MD5_GG    (c,d,a,b,x11, 2,14    ); MD5_GG  (b,c,d,a, x0, 3,20    );
	MD5_GG  (a,b,c,d, x5, 4, 5    ); MD5_GG  (d,a,b,c,x10, 5, 9    ); MD5_GG    (c,d,a,b,x15, 6,14    ); MD5_GG  (b,c,d,a, x4, 7,20    );
	MD5_GG  (a,b,c,d, x9, 8, 5    ); MD5_GG  (d,a,b,c,x14, 9, 9    ); MD5_GG    (c,d,a,b, x3,10,14    ); MD5_GG  (b,c,d,a, x8,11,20    );
	MD5_GG  (a,b,c,d,x13,12, 5    ); MD5_GG  (d,a,b,c, x2,13, 9    ); MD5_GG    (c,d,a,b, x7,14,14    ); MD5_GG  (b,c,d,a,x12,15,20    );

	register __m128i t[__MD5_INTERLACED__];
	MD5_HH_E(a,b,c,d, x5, 0, 4,t); MD5_HH_O(d,a,b,c, x8, 1,11,t); MD5_HH16_E(c,d,a,b,x11, 2,16,t); MD5_HH_O(b,c,d,a,x14, 3,23,t);
	MD5_HH_E(a,b,c,d, x1, 4, 4,t); MD5_HH_O(d,a,b,c, x4, 5,11,t); MD5_HH16_E(c,d,a,b, x7, 6,16,t); MD5_HH_O(b,c,d,a,x10, 7,23,t);
	MD5_HH_E(a,b,c,d,x13, 8, 4,t); MD5_HH_O(d,a,b,c, x0, 9,11,t); MD5_HH16_E(c,d,a,b, x3,10,16,t); MD5_HH_O(b,c,d,a, x6,11,23,t);
	MD5_HH_E(a,b,c,d, x9,12, 4,t); MD5_HH_O(d,a,b,c,x12,13,11,t); MD5_HH16_E(c,d,a,b,x15,14,16,t); MD5_HH_O(b,c,d,a, x2,15,23,t);

	register __m128i n = _mm_set_epi32(0xffffffff,0xffffffff,0xffffffff,0xffffffff);
	MD5_II  (a,b,c,d, x0, 0, 6,n); MD5_II  (d,a,b,c, x7, 1,10,n); MD5_II    (c,d,a,b,x14, 2,15,n); MD5_II  (b,c,d,a, x5, 3,21,n);
	MD5_II  (a,b,c,d,x12, 4, 6,n); MD5_II  (d,a,b,c, x3, 5,10,n); MD5_II    (c,d,a,b,x10, 6,15,n); MD5_II  (b,c,d,a, x1, 7,21,n);
	MD5_II  (a,b,c,d, x8, 8, 6,n); MD5_II  (d,a,b,c,x15, 9,10,n); MD5_II    (c,d,a,b, x6,10,15,n); MD5_II  (b,c,d,a,x13,11,21,n);
	MD5_II  (a,b,c,d, x4,12, 6,n); MD5_II  (d,a,b,c,x11,13,10,n); MD5_II    (c,d,a,b, x2,14,15,n); MD5_II  (b,c,d,a, x9,15,21,n);

	if (__MD5_INTERLACED__ > 0) { hash[0 * __MD5_INTERLACED__ + 0] = _mm_add_epi32(a[0], LOAD_CONST_MD5_A); }
	if (__MD5_INTERLACED__ > 1) { hash[0 * __MD5_INTERLACED__ + 1] = _mm_add_epi32(a[1], LOAD_CONST_MD5_A); }
	if (__MD5_INTERLACED__ > 2) { hash[0 * __MD5_INTERLACED__ + 2] = _mm_add_epi32(a[2], LOAD_CONST_MD5_A); }
	if (__MD5_INTERLACED__ > 3) { hash[0 * __MD5_INTERLACED__ + 3] = _mm_add_epi32(a[3], LOAD_CONST_MD5_A); }
	if (__MD5_INTERLACED__ > 0) { hash[1 * __MD5_INTERLACED__ + 0] = _mm_add_epi32(b[0], LOAD_CONST_MD5_B); }
	if (__MD5_INTERLACED__ > 1) { hash[1 * __MD5_INTERLACED__ + 1] = _mm_add_epi32(b[1], LOAD_CONST_MD5_B); }
	if (__MD5_INTERLACED__ > 2) { hash[1 * __MD5_INTERLACED__ + 2] = _mm_add_epi32(b[2], LOAD_CONST_MD5_B); }
	if (__MD5_INTERLACED__ > 3) { hash[1 * __MD5_INTERLACED__ + 3] = _mm_add_epi32(b[3], LOAD_CONST_MD5_B); }
	if (__MD5_INTERLACED__ > 0) { hash[2 * __MD5_INTERLACED__ + 0] = _mm_add_epi32(c[0], LOAD_CONST_MD5_C); }
	if (__MD5_INTERLACED__ > 1) { hash[2 * __MD5_INTERLACED__ + 1] = _mm_add_epi32(c[1], LOAD_CONST_MD5_C); }
	if (__MD5_INTERLACED__ > 2) { hash[2 * __MD5_INTERLACED__ + 2] = _mm_add_epi32(c[2], LOAD_CONST_MD5_C); }
	if (__MD5_INTERLACED__ > 3) { hash[2 * __MD5_INTERLACED__ + 3] = _mm_add_epi32(c[3], LOAD_CONST_MD5_C); }
	if (__MD5_INTERLACED__ > 0) { hash[3 * __MD5_INTERLACED__ + 0] = _mm_add_epi32(d[0], LOAD_CONST_MD5_D); }
	if (__MD5_INTERLACED__ > 1) { hash[3 * __MD5_INTERLACED__ + 1] = _mm_add_epi32(d[1], LOAD_CONST_MD5_D); }
	if (__MD5_INTERLACED__ > 2) { hash[3 * __MD5_INTERLACED__ + 2] = _mm_add_epi32(d[2], LOAD_CONST_MD5_D); }
	if (__MD5_INTERLACED__ > 3) { hash[3 * __MD5_INTERLACED__ + 3] = _mm_add_epi32(d[3], LOAD_CONST_MD5_D); }
}
#undef ROT
#undef ROT16

#ifdef XOP_CODE
#undef MD5_FF_
#undef MD5_GG_
#define MD5_FF_(a,b,c,d,x,i,j,s)  if (__MD5_INTERLACED__ > j) { a[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(a[j], _mm_cmov_si128(c[j], d[j], b[j])), x(j)), LOAD_CONST_MD5_FF##i); a[j] = _mm_add_epi32(ROT(a[j], s), b[j]); }
#define MD5_GG_(a,b,c,d,x,i,j,s)  if (__MD5_INTERLACED__ > j) { a[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(a[j], _mm_cmov_si128(b[j], c[j], d[j])), x(j)), LOAD_CONST_MD5_GG##i); a[j] = _mm_add_epi32(ROT(a[j], s), b[j]); }
#define ROT(n,s) _mm_roti_epi32(n, s)
#define ROT16(n) _mm_roti_epi32(n, 16)
__inline__ void md5Func_XOP(__m128i *pw, __m128i *hash)
{
	__m128i a[__MD5_INTERLACED__], b[__MD5_INTERLACED__], c[__MD5_INTERLACED__], d[__MD5_INTERLACED__];

	if (__MD5_INTERLACED__ > 0) { a[0] = LOAD_CONST_MD5_A; }
	if (__MD5_INTERLACED__ > 1) { a[1] = LOAD_CONST_MD5_A; }
	if (__MD5_INTERLACED__ > 2) { a[2] = LOAD_CONST_MD5_A; }
	if (__MD5_INTERLACED__ > 3) { a[3] = LOAD_CONST_MD5_A; }
	if (__MD5_INTERLACED__ > 0) { b[0] = LOAD_CONST_MD5_B; }
	if (__MD5_INTERLACED__ > 1) { b[1] = LOAD_CONST_MD5_B; }
	if (__MD5_INTERLACED__ > 2) { b[2] = LOAD_CONST_MD5_B; }
	if (__MD5_INTERLACED__ > 3) { b[3] = LOAD_CONST_MD5_B; }
	if (__MD5_INTERLACED__ > 0) { c[0] = LOAD_CONST_MD5_C; }
	if (__MD5_INTERLACED__ > 1) { c[1] = LOAD_CONST_MD5_C; }
	if (__MD5_INTERLACED__ > 2) { c[2] = LOAD_CONST_MD5_C; }
	if (__MD5_INTERLACED__ > 3) { c[3] = LOAD_CONST_MD5_C; }
	if (__MD5_INTERLACED__ > 0) { d[0] = LOAD_CONST_MD5_D; }
	if (__MD5_INTERLACED__ > 1) { d[1] = LOAD_CONST_MD5_D; }
	if (__MD5_INTERLACED__ > 2) { d[2] = LOAD_CONST_MD5_D; }
	if (__MD5_INTERLACED__ > 3) { d[3] = LOAD_CONST_MD5_D; }

	if (__MD5_INTERLACED__ > 0) { x14(0) = _mm_slli_epi32(x14(0), 3); }
	if (__MD5_INTERLACED__ > 1) { x14(1) = _mm_slli_epi32(x14(1), 3); }
	if (__MD5_INTERLACED__ > 2) { x14(2) = _mm_slli_epi32(x14(2), 3); }
	if (__MD5_INTERLACED__ > 3) { x14(3) = _mm_slli_epi32(x14(3), 3); }

	MD5_FF  (a,b,c,d, x0, 0, 7  ); MD5_FF  (d,a,b,c, x1, 1,12  ); MD5_FF    (c,d,a,b, x2, 2,17  ); MD5_FF  (b,c,d,a, x3, 3,22  );
	MD5_FF  (a,b,c,d, x4, 4, 7  ); MD5_FF  (d,a,b,c, x5, 5,12  ); MD5_FF    (c,d,a,b, x6, 6,17  ); MD5_FF  (b,c,d,a, x7, 7,22  );
	MD5_FF  (a,b,c,d, x8, 8, 7  ); MD5_FF  (d,a,b,c, x9, 9,12  ); MD5_FF    (c,d,a,b,x10,10,17  ); MD5_FF  (b,c,d,a,x11,11,22  );
	MD5_FF  (a,b,c,d,x12,12, 7  ); MD5_FF  (d,a,b,c,x13,13,12  ); MD5_FF    (c,d,a,b,x14,14,17  ); MD5_FF  (b,c,d,a,x15,15,22  );

	MD5_GG  (a,b,c,d, x1, 0, 5  ); MD5_GG  (d,a,b,c, x6, 1, 9  ); MD5_GG    (c,d,a,b,x11, 2,14  ); MD5_GG  (b,c,d,a, x0, 3,20  );
	MD5_GG  (a,b,c,d, x5, 4, 5  ); MD5_GG  (d,a,b,c,x10, 5, 9  ); MD5_GG    (c,d,a,b,x15, 6,14  ); MD5_GG  (b,c,d,a, x4, 7,20  );
	MD5_GG  (a,b,c,d, x9, 8, 5  ); MD5_GG  (d,a,b,c,x14, 9, 9  ); MD5_GG    (c,d,a,b, x3,10,14  ); MD5_GG  (b,c,d,a, x8,11,20  );
	MD5_GG  (a,b,c,d,x13,12, 5  ); MD5_GG  (d,a,b,c, x2,13, 9  ); MD5_GG    (c,d,a,b, x7,14,14  ); MD5_GG  (b,c,d,a,x12,15,20  );

	register __m128i t[__MD5_INTERLACED__];
	MD5_HH_E(a,b,c,d, x5, 0, 4,t); MD5_HH_O(d,a,b,c, x8, 1,11,t); MD5_HH16_E(c,d,a,b,x11, 2,16,t); MD5_HH_O(b,c,d,a,x14, 3,23,t);
	MD5_HH_E(a,b,c,d, x1, 4, 4,t); MD5_HH_O(d,a,b,c, x4, 5,11,t); MD5_HH16_E(c,d,a,b, x7, 6,16,t); MD5_HH_O(b,c,d,a,x10, 7,23,t);
	MD5_HH_E(a,b,c,d,x13, 8, 4,t); MD5_HH_O(d,a,b,c, x0, 9,11,t); MD5_HH16_E(c,d,a,b, x3,10,16,t); MD5_HH_O(b,c,d,a, x6,11,23,t);
	MD5_HH_E(a,b,c,d, x9,12, 4,t); MD5_HH_O(d,a,b,c,x12,13,11,t); MD5_HH16_E(c,d,a,b,x15,14,16,t); MD5_HH_O(b,c,d,a, x2,15,23,t);

	register __m128i n = _mm_set_epi32(0xffffffff,0xffffffff,0xffffffff,0xffffffff);
	MD5_II  (a,b,c,d, x0, 0, 6,n); MD5_II  (d,a,b,c, x7, 1,10,n); MD5_II    (c,d,a,b,x14, 2,15,n); MD5_II  (b,c,d,a, x5, 3,21,n);
	MD5_II  (a,b,c,d,x12, 4, 6,n); MD5_II  (d,a,b,c, x3, 5,10,n); MD5_II    (c,d,a,b,x10, 6,15,n); MD5_II  (b,c,d,a, x1, 7,21,n);
	MD5_II  (a,b,c,d, x8, 8, 6,n); MD5_II  (d,a,b,c,x15, 9,10,n); MD5_II    (c,d,a,b, x6,10,15,n); MD5_II  (b,c,d,a,x13,11,21,n);
	MD5_II  (a,b,c,d, x4,12, 6,n); MD5_II  (d,a,b,c,x11,13,10,n); MD5_II    (c,d,a,b, x2,14,15,n); MD5_II  (b,c,d,a, x9,15,21,n);

	if (__MD5_INTERLACED__ > 0) { hash[0 * __MD5_INTERLACED__ + 0] = _mm_add_epi32(a[0], LOAD_CONST_MD5_A); }
	if (__MD5_INTERLACED__ > 1) { hash[0 * __MD5_INTERLACED__ + 1] = _mm_add_epi32(a[1], LOAD_CONST_MD5_A); }
	if (__MD5_INTERLACED__ > 2) { hash[0 * __MD5_INTERLACED__ + 2] = _mm_add_epi32(a[2], LOAD_CONST_MD5_A); }
	if (__MD5_INTERLACED__ > 3) { hash[0 * __MD5_INTERLACED__ + 3] = _mm_add_epi32(a[3], LOAD_CONST_MD5_A); }
	if (__MD5_INTERLACED__ > 0) { hash[1 * __MD5_INTERLACED__ + 0] = _mm_add_epi32(b[0], LOAD_CONST_MD5_B); }
	if (__MD5_INTERLACED__ > 1) { hash[1 * __MD5_INTERLACED__ + 1] = _mm_add_epi32(b[1], LOAD_CONST_MD5_B); }
	if (__MD5_INTERLACED__ > 2) { hash[1 * __MD5_INTERLACED__ + 2] = _mm_add_epi32(b[2], LOAD_CONST_MD5_B); }
	if (__MD5_INTERLACED__ > 3) { hash[1 * __MD5_INTERLACED__ + 3] = _mm_add_epi32(b[3], LOAD_CONST_MD5_B); }
	if (__MD5_INTERLACED__ > 0) { hash[2 * __MD5_INTERLACED__ + 0] = _mm_add_epi32(c[0], LOAD_CONST_MD5_C); }
	if (__MD5_INTERLACED__ > 1) { hash[2 * __MD5_INTERLACED__ + 1] = _mm_add_epi32(c[1], LOAD_CONST_MD5_C); }
	if (__MD5_INTERLACED__ > 2) { hash[2 * __MD5_INTERLACED__ + 2] = _mm_add_epi32(c[2], LOAD_CONST_MD5_C); }
	if (__MD5_INTERLACED__ > 3) { hash[2 * __MD5_INTERLACED__ + 3] = _mm_add_epi32(c[3], LOAD_CONST_MD5_C); }
	if (__MD5_INTERLACED__ > 0) { hash[3 * __MD5_INTERLACED__ + 0] = _mm_add_epi32(d[0], LOAD_CONST_MD5_D); }
	if (__MD5_INTERLACED__ > 1) { hash[3 * __MD5_INTERLACED__ + 1] = _mm_add_epi32(d[1], LOAD_CONST_MD5_D); }
	if (__MD5_INTERLACED__ > 2) { hash[3 * __MD5_INTERLACED__ + 2] = _mm_add_epi32(d[2], LOAD_CONST_MD5_D); }
	if (__MD5_INTERLACED__ > 3) { hash[3 * __MD5_INTERLACED__ + 3] = _mm_add_epi32(d[3], LOAD_CONST_MD5_D); }
}
#undef ROT
#undef ROT16
#endif

#undef x0
#undef x1
#undef x2
#undef x3
#undef x4
#undef x5
#undef x6
#undef x7
#undef x8
#undef x9
#undef x10
#undef x11
#undef x12
#undef x13
#undef x14
#undef x15

#undef ZERO

#undef MD5_FF
#undef MD5_GG
#undef MD5_HH
#undef MD5_HH_E
#undef MD5_HH_O
#undef MD5_HH16
#undef MD5_HH16_E
#undef MD5_II

#undef MD5_FF_
#undef MD5_GG_
#undef MD5_HH_
#undef MD5_HH_E_
#undef MD5_HH_O_
#undef MD5_HH16_
#undef MD5_HH16_E_
#undef MD5_II_

#undef LOAD_CONST_MD5_A
#undef LOAD_CONST_MD5_B
#undef LOAD_CONST_MD5_C
#undef LOAD_CONST_MD5_D

#undef LOAD_CONST_MD5_FF0
#undef LOAD_CONST_MD5_FF1
#undef LOAD_CONST_MD5_FF2
#undef LOAD_CONST_MD5_FF3
#undef LOAD_CONST_MD5_FF4
#undef LOAD_CONST_MD5_FF5
#undef LOAD_CONST_MD5_FF6
#undef LOAD_CONST_MD5_FF7
#undef LOAD_CONST_MD5_FF8
#undef LOAD_CONST_MD5_FF9
#undef LOAD_CONST_MD5_FF10
#undef LOAD_CONST_MD5_FF11
#undef LOAD_CONST_MD5_FF12
#undef LOAD_CONST_MD5_FF13
#undef LOAD_CONST_MD5_FF14
#undef LOAD_CONST_MD5_FF15

#undef LOAD_CONST_MD5_GG0
#undef LOAD_CONST_MD5_GG1
#undef LOAD_CONST_MD5_GG2
#undef LOAD_CONST_MD5_GG3
#undef LOAD_CONST_MD5_GG4
#undef LOAD_CONST_MD5_GG5
#undef LOAD_CONST_MD5_GG6
#undef LOAD_CONST_MD5_GG7
#undef LOAD_CONST_MD5_GG8
#undef LOAD_CONST_MD5_GG9
#undef LOAD_CONST_MD5_GG10
#undef LOAD_CONST_MD5_GG11
#undef LOAD_CONST_MD5_GG12
#undef LOAD_CONST_MD5_GG13
#undef LOAD_CONST_MD5_GG14
#undef LOAD_CONST_MD5_GG15

#undef LOAD_CONST_MD5_HH0
#undef LOAD_CONST_MD5_HH1
#undef LOAD_CONST_MD5_HH2
#undef LOAD_CONST_MD5_HH3
#undef LOAD_CONST_MD5_HH4
#undef LOAD_CONST_MD5_HH5
#undef LOAD_CONST_MD5_HH6
#undef LOAD_CONST_MD5_HH7
#undef LOAD_CONST_MD5_HH8
#undef LOAD_CONST_MD5_HH9
#undef LOAD_CONST_MD5_HH10
#undef LOAD_CONST_MD5_HH11
#undef LOAD_CONST_MD5_HH12
#undef LOAD_CONST_MD5_HH13
#undef LOAD_CONST_MD5_HH14
#undef LOAD_CONST_MD5_HH15

#undef LOAD_CONST_MD5_II0
#undef LOAD_CONST_MD5_II1
#undef LOAD_CONST_MD5_II2
#undef LOAD_CONST_MD5_II3
#undef LOAD_CONST_MD5_II4
#undef LOAD_CONST_MD5_II5
#undef LOAD_CONST_MD5_II6
#undef LOAD_CONST_MD5_II7
#undef LOAD_CONST_MD5_II8
#undef LOAD_CONST_MD5_II9
#undef LOAD_CONST_MD5_II10
#undef LOAD_CONST_MD5_II11
#undef LOAD_CONST_MD5_II12
#undef LOAD_CONST_MD5_II13
#undef LOAD_CONST_MD5_II14
#undef LOAD_CONST_MD5_II15

#endif
#endif
