/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "redux.h"

Redux::~Redux()
{
}

Redux::Redux(const char *hashName, uint32_t maxPwLen)
{
	hashAlgorithm ha;

	m_hashFunc = NULL;
	HashFactory::getHashInstance(hashName, 0, maxPwLen, ha);
	if (ha.func == NULL)
	{
		reportError("No suitable hashing algorithm found.");
	}
	else
	{
		m_maxPwLen = maxPwLen;
		m_characterEncoding = ha.characterEncoding;
		m_hashFunc = ha.func;
		m_hashLen = ha.len;
		m_optimalNumIndexes = ha.hashesPerCall;
		m_pwBlocks = ha.numPwBlocks;
		m_vecSize = (ha.hashesPerCall + 3) / 4;
	}
}

void Redux::cleanUp()
{
	m_hashFunc = NULL;
}

void Redux::copy(const Redux &r)
{
	m_hashFunc = NULL;
	if (r.m_hashFunc != NULL)
	{
		m_keySpace          = r.m_keySpace;
		m_hashFunc          = r.m_hashFunc;
		m_hashLen           = r.m_hashLen;
		m_optimalNumIndexes = r.m_optimalNumIndexes;
		m_maxPwLen          = r.m_maxPwLen;
		m_characterEncoding = r.m_characterEncoding;
		m_pwBlocks          = r.m_pwBlocks;
		m_vecSize           = r.m_vecSize;
	}
}
