/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef REDUX_FACTORY_H
#define REDUX_FACTORY_H

#include <string>
#include <list>
#include "common.h"
#include "redux.h"
#include "keyspace.h"

/*
	div  RCrack's method
	fpm  Fixed point multiplication method
	lut  Look up table method
	key  Next pw = hash, method
*/
typedef Redux *(*CREATE_REDUX)(const keySpace &ks, const char *hashName);
typedef Redux *(*CREATE_REDUX_GPU)(const keySpace &ks, const char *hashName, int gpu);

struct reduxInfo
{
	CREATE_REDUX createRedux;
	CREATE_REDUX_GPU createReduxGpu;
	std::string name, comment;
	uint32_t charEncodings;
};

// You must call ReduxFactory::getReduxInstance() for each thread!
// Since some Redux classes are not thread safe to prevent unnecessary creation and deletion
// of memory every function call. They instead allocate that memory on class creation.
class ReduxFactory
{
public:
	static Redux*   getReduxInstance(const char *name, const keySpace &ks, int gpu, const char *hashName);
	static uint32_t registerRedux(const char *name, const char *comment, uint32_t charEncodings, CREATE_REDUX createRedux, CREATE_REDUX_GPU createReduxGpu = NULL);
	static void     printReduxes();

private:
	ReduxFactory() {}
	static std::list<reduxInfo*> *s_reduxes;
};

#endif
