/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "writefactory.h"

using namespace std;

map<string,writeInfo*> *WriteFactory::s_writes = NULL;

Write* WriteFactory::getWriteInstance(const char *name, const char *folder, const char *fileNameComment, const char *reduxName, const char *hashName, const char *salt, const char *keySpaceName, uint32_t chainLength, uint32_t tableIndex)
{
	Write *ret = NULL;

	if (s_writes != NULL)
	{
		map<string,writeInfo*>::iterator it;

		it = s_writes->find(name);
		if (it != s_writes->end())
		{
			ret = it->second->createWrite(folder, fileNameComment, reduxName, hashName, salt, keySpaceName, chainLength, tableIndex);
		}
	}
	return ret;
}

uint32_t WriteFactory::registerWrite(const char *name, const char *comment, CREATE_WRITE createWrite)
{
	writeInfo *wi = new writeInfo;

	wi->createWrite = createWrite;
	wi->comment     = comment;
	if (s_writes == NULL)
	{
		s_writes = new map<string,writeInfo*>;
	}
	(*s_writes)[name] = wi;
	return 1;
}

void WriteFactory::printWrites()
{
	map<string,writeInfo*>::iterator it;
	char *spaces;
	uint32_t maxLen = 0;

	if (s_writes != NULL)
	{
		// Find max name length
		for (it = s_writes->begin(); it != s_writes->end(); ++it)
		{
			if (maxLen < it->first.length())
				maxLen = it->first.length();
		}

		// Fill spaces with max name length number of spaces ' '
		spaces = new char[maxLen + 1];
		for (uint32_t a = 0; a < maxLen; a++)
		{
			spaces[a] = ' ';
		}
		spaces[maxLen] = 0;

		// Print
		for (it = s_writes->begin(); it != s_writes->end(); ++it)
		{
			printf(" * %s:%s %s\n",
				it->first.c_str(), spaces + (maxLen - it->first.length()),
				it->second->comment.c_str());
		}
		delete [] spaces;
	}
	printf("\n");
}
