/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef WRITE_FACTORY_H
#define WRITE_FACTORY_H

#include <string>
#include <map>
#include "common.h"
#include "write.h"
#include "keyspace.h"

typedef Write *(*CREATE_WRITE)(const char *folder, const char *fileNameComment, const char *reduxName, const char *hashName, const char *salt, const char *keySpaceName, uint32_t chainLength, uint32_t tableIndex);

struct writeInfo
{
	CREATE_WRITE createWrite;
	std::string comment;
};

class WriteFactory
{
public:
	static Write*   getWriteInstance(const char *name, const char *folder, const char *fileNameComment, const char *reduxName, const char *hashName, const char *salt, const char *keySpaceName, uint32_t chainLength, uint32_t tableIndex);
	static uint32_t registerWrite(const char *name, const char *comment, CREATE_WRITE createWrite);
	static void     printWrites();

private:
	WriteFactory() {}
	static std::map<std::string,writeInfo*> *s_writes;
};

#endif
